'''
BSD 3-Clause License

Copyright (c) 2021, Qing Yu
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its
   contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
'''

import geopandas as gpd
import pandas as pd
from .grids import GPS_to_grid, grid_to_centre


def odagg_grid(oddata, params, col=['slon', 'slat', 'elon', 'elat'],
               arrow=False, **kwargs):
    '''
    Aggregate the OD matrix and generate the grid geometry.
    The input is the OD matrix (each row represents a trip).
    The OD will assigned to grids and then aggregated in the form of
    GeoDataFrame.

    Parameters
    -------
    oddata : DataFrame
        OD data
    col : List
        The column of the origin/destination location,[‘slon’,’slat’,
        ’elon’,’elat’]. The default weight is 1 for each column.
        You can also add the weight parameter, for example, [‘slon’,
        ’slat’,’elon’,’elat’,’count’].
    params : List
        Gridding parameters (lonStart,latStart,deltaLon,deltaLat),
        lonStart and latStart are the lower-left coordinates,
        deltaLon, deltaLat are the length and width of a single grid
    arrow : bool
        Whether the generated OD geographic line contains arrows

    Returns
    -------
    oddata1 : GeoDataFrame
        GeoDataFrame of OD after aggregation
    '''
    if len(col) == 4:
        [slon, slat, elon, elat] = col
        count = 'count'
    if len(col) == 5:
        [slon, slat, elon, elat, count] = col
    oddata['SLONCOL'], oddata['SLATCOL'] = GPS_to_grid(
        oddata[slon], oddata[slat], params)
    oddata['ELONCOL'], oddata['ELATCOL'] = GPS_to_grid(
        oddata[elon], oddata[elat], params)
    if len(col) == 4:
        oddata[count] = 1
    oddata_agg = oddata.groupby(['SLONCOL', 'SLATCOL', 'ELONCOL', 'ELATCOL'])[
        count].sum().reset_index()
    oddata_agg['SHBLON'], oddata_agg['SHBLAT'] = grid_to_centre(
        [oddata_agg['SLONCOL'], oddata_agg['SLATCOL']], params)
    oddata_agg['EHBLON'], oddata_agg['EHBLAT'] = grid_to_centre(
        [oddata_agg['ELONCOL'], oddata_agg['ELATCOL']], params)
    from shapely.geometry import LineString
    if arrow:
        oddata_agg['geometry'] = oddata_agg.apply(lambda r: tolinewitharrow(
            r['SHBLON'], r['SHBLAT'], r['EHBLON'], r['EHBLAT'],
            **kwargs), axis=1)
    else:
        oddata_agg['geometry'] = oddata_agg.apply(lambda r: LineString(
            [[r['SHBLON'], r['SHBLAT']], [r['EHBLON'], r['EHBLAT']]]), axis=1)
    oddata_agg = gpd.GeoDataFrame(oddata_agg)
    oddata_agg = oddata_agg.sort_values(by=count)
    return oddata_agg


def odagg_shape(oddata, shape, col=['slon', 'slat', 'elon', 'elat'],
                params=None, round_accuracy=6, arrow=False, **kwargs):
    '''
    Generate the OD aggregation results and the corresponding geometry.
    The input is the OD data (each row represents a trip). The OD will
    assigned to grids and then aggregated in the form of GeoDataFrame.

    Parameters
    -------
    oddata : DataFrame
        OD data
    shape : GeoDataFrame
        GeoDataFrame of the target traffic zone
    col : List
        The column of the origin/destination location,[‘slon’,’slat’,’elon’
        ,’elat’]. The default weight is 1 for each column. You can also add
        the weight parameter, for example, [‘slon’,’slat’,’elon’,’elat’,
        ’count’].
    params : List (optional)
        Gridding parameters (lonStart,latStart,deltaLon,deltaLat), lonStart
        and latStart are the lower-left coordinates, deltaLon, deltaLat are
        the length and width of a single grid
        If availabel, After the data gridding, the traffic zone will be matched
        based on the grid center. If not available, then the matching will be
        processed based on longitude and latitude. When the number of data
        items is large, the matching efficiency will be improved greatly thanks
        to data gridding.
    round_accuracy : number
        The number of decimal for latitude and longitude when implementing
        aggregation
    arrow : bool
        Whether the generated OD geographic line contains arrows

    Returns
    -------
    oddata1 : GeoDataFrame
        GeoDataFrame of OD after aggregation
    '''
    if len(col) == 4:
        [slon, slat, elon, elat] = col
        count = 'count'
    if len(col) == 5:
        [slon, slat, elon, elat, count] = col
    shape_1 = shape.copy()
    shape['x'] = shape.centroid.x
    shape['y'] = shape.centroid.y
    shape = shape[['x', 'y', 'geometry']]
    if params:
        oddata['SLONCOL'], oddata['SLATCOL'] = GPS_to_grid(
            oddata[slon], oddata[slat], params)
        oddata['ELONCOL'], oddata['ELATCOL'] = GPS_to_grid(
            oddata[elon], oddata[elat], params)
        if len(col) == 4:
            oddata[count] = 1
        oddata_agg = oddata.groupby([
            'SLONCOL', 'SLATCOL', 'ELONCOL', 'ELATCOL'])[
                count].sum().reset_index()
        oddata_agg['SHBLON'], oddata_agg['SHBLAT'] = grid_to_centre(
            [oddata_agg['SLONCOL'], oddata_agg['SLATCOL']], params)
        oddata_agg['EHBLON'], oddata_agg['EHBLAT'] = grid_to_centre(
            [oddata_agg['ELONCOL'], oddata_agg['ELATCOL']], params)
        a = oddata_agg[['SHBLON', 'SHBLAT']]
        b = oddata_agg[['EHBLON', 'EHBLAT']]
        a.columns = ['lon', 'lat']
        b.columns = ['lon', 'lat']
        c = pd.concat([a, b]).drop_duplicates()
        d = c[['lon', 'lat']].drop_duplicates()
        d['geometry'] = gpd.points_from_xy(d['lon'], d['lat'])
        d = gpd.GeoDataFrame(d)
        d = gpd.sjoin(d, shape)
        c = pd.merge(c, d)
        c = c[['lon', 'lat', 'index_right', 'x', 'y']]
        c.columns = ['SHBLON', 'SHBLAT', 'sindex', 'sx', 'sy']
        oddata_agg = pd.merge(oddata_agg, c)
        c.columns = ['EHBLON', 'EHBLAT', 'eindex', 'ex', 'ey']
        oddata_agg = pd.merge(oddata_agg, c)
        oddata_agg = oddata_agg.groupby([
            'sindex', 'sx', 'sy', 'eindex', 'ex', 'ey'])[
            count].sum().reset_index()
    else:
        a = oddata[[slon, slat]]
        b = oddata[[elon, elat]]
        a.columns = ['lon', 'lat']
        b.columns = ['lon', 'lat']
        c = pd.concat([a, b]).drop_duplicates()
        c['lon_simple'] = c['lon'].round(round_accuracy)
        c['lat_simple'] = c['lat'].round(round_accuracy)
        d = c[['lon_simple', 'lat_simple']].drop_duplicates()
        d['geometry'] = gpd.points_from_xy(d['lon_simple'], d['lat_simple'])
        d = gpd.GeoDataFrame(d)
        d = gpd.sjoin(d, shape)
        c = pd.merge(c, d)
        c = c[['lon', 'lat', 'index_right', 'x', 'y']]
        c.columns = ['slon', 'slat', 'sindex', 'sx', 'sy']
        oddata = pd.merge(oddata, c)
        c.columns = ['elon', 'elat', 'eindex', 'ex', 'ey']
        oddata = pd.merge(oddata, c)
        if len(col) == 4:
            oddata[count] = 1
        oddata_agg = oddata.groupby([
            'sindex', 'sx', 'sy', 'eindex', 'ex', 'ey'])[
            count].sum().reset_index()
    from shapely.geometry import LineString
    if arrow:
        oddata_agg['geometry'] = oddata_agg.apply(lambda r: tolinewitharrow(
            r['sx'], r['sy'], r['ex'], r['ey'], **kwargs), axis=1)
    else:
        oddata_agg['geometry'] = oddata_agg.apply(lambda r: LineString(
            [[r['sx'], r['sy']], [r['ex'], r['ey']]]), axis=1)
    oddata_agg = gpd.GeoDataFrame(oddata_agg)
    oddata_agg = oddata_agg[['sindex', 'eindex', count, 'geometry']]
    oddata_agg = pd.merge(oddata_agg, shape_1.reset_index().rename(
        columns={'index': 'sindex'}).drop('geometry', axis=1), on='sindex')
    oddata_agg = pd.merge(oddata_agg, shape_1.reset_index().rename(
        columns={'index': 'eindex'}).drop('geometry', axis=1), on='eindex')
    oddata_agg = oddata_agg.sort_values(by=count)
    return oddata_agg


def tolinewitharrow(x1, y1, x2, y2, theta=20, length=0.1, pos=0.8):
    '''
    Input start and end coords，Returns LineString with arrow

    Parameters
    -------
    x1,y1,x2,y2 : float
        xy coords of start and end coords
    theta : float
        Angle of arrow
    length : float
        The length ratio of the arrow to the original line. For example, if the
        length of the original line is 1 and the length is set to 0.3, the
        arrow size is 0.3
    pos : float
        Position of arrow, 0 at the start point, 1 at the end point

    Returns
    -------
    Line : MultiLineString
        OD LineString with arrow
    '''
    import numpy as np
    from shapely.geometry import MultiLineString
    l_main = [[x1, y1], [x2, y2]]
    p1, p2 = (1-pos)*x1+pos*x2, (1-pos)*y1+pos*y2
    R = np.array([[np.cos(np.radians(theta)), -np.sin(np.radians(theta))],
                  [np.sin(np.radians(theta)), np.cos(np.radians(theta))]])
    l1 = np.dot(R, np.array([[x1-x2, y1-y2]]).T).T[0] * \
        length+np.array([p1, p2]).T
    l1 = [list(l1), [p1, p2]]
    R = np.array([[np.cos(np.radians(-theta)), -np.sin(np.radians(-theta))],
                  [np.sin(np.radians(-theta)), np.cos(np.radians(-theta))]])
    l2 = np.dot(R, np.array([[x1-x2, y1-y2]]).T).T[0] * \
        length+np.array([p1, p2]).T
    l2 = [list(l2), [p1, p2]]
    return MultiLineString([l_main, l1, l2])



'''
BSD 3-Clause License

Copyright (c) 2021, Qing Yu
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its
   contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
'''

import pandas as pd
import numpy as np
from .grids import GPS_to_grid, grid_to_centre


def mobile_stay_move(data, params,
                     col=['ID', 'dataTime', 'longitude', 'latitude'],
                     activitytime=1800):
    '''
    Input trajectory data and gridding parameters, identify stay and move

    Parameters
    ----------------
    data : DataFrame
        trajectory data
    params : List
        gridding parameters
    col : List
        The column name, in the order of ['ID','dataTime','longitude',
        'latitude']
    activitytime : Number
        How much time to regard as activity

    Returns
    ----------------
    stay : DataFrame
        stay information
    move : DataFrame
        move information
    '''
    uid, timecol, lon, lat = col
    # Identify stay
    data = data.sort_values(by = col[:2])
    stay = data.copy()
    stay = stay.rename(columns={lon: 'lon', lat: 'lat', timecol: 'stime'})
    stay['stime'] = pd.to_datetime(stay['stime'])
    stay['LONCOL'], stay['LATCOL'] = GPS_to_grid(
        stay['lon'], stay['lat'], params)
    # Number the status
    stay['status_id'] = ((stay['LONCOL'] != stay['LONCOL'].shift()) |
                         (stay['LATCOL'] != stay['LATCOL'].shift()) |
                         (stay[uid] != stay[uid].shift())).astype(int)
    stay['status_id'] = stay.groupby([uid])['status_id'].cumsum()
    stay = stay.drop_duplicates(
        subset=[uid, 'status_id'], keep='first').copy()
    stay['etime'] = stay['stime'].shift(-1)
    stay = stay[stay[uid] == stay[uid].shift(-1)].copy()
    # Remove the duration shorter than given activitytime
    stay['duration'] = (pd.to_datetime(stay['etime']) -
                        pd.to_datetime(stay['stime'])).dt.total_seconds()
    stay = stay[stay['duration'] >= activitytime].copy()
    # Renumber the status
    stay['status_id'] = ((stay['LONCOL'] != stay['LONCOL'].shift()) |
                         (stay['LATCOL'] != stay['LATCOL'].shift()) |
                         (stay[uid] != stay[uid].shift())).astype(int)
    stay['status_id'] = stay.groupby([uid])['status_id'].cumsum()
    stay_stime = stay.drop_duplicates(subset=[uid, 'status_id'], keep='first')[
        [uid, 'stime', 'LONCOL', 'LATCOL', 'status_id']]
    stay_etime = stay.drop_duplicates(subset=[uid, 'status_id'], keep='last')[
        [uid, 'etime', 'LONCOL', 'LATCOL', 'status_id']]
    stay = pd.merge(stay_stime, stay_etime, how='left').drop(
        'status_id', axis=1)
    stay['lon'], stay['lat'] = grid_to_centre(
        [stay['LONCOL'], stay['LATCOL']], params)
    stay['duration'] = (pd.to_datetime(stay['etime']) -
                        pd.to_datetime(stay['stime'])).dt.total_seconds()
    # Identify move
    move = stay.copy()
    move['stime_next'] = move['stime'].shift(-1)
    move['elon'] = move['lon'].shift(-1)
    move['elat'] = move['lat'].shift(-1)
    move['ELONCOL'] = move['LONCOL'].shift(-1)
    move['ELATCOL'] = move['LATCOL'].shift(-1)
    move[uid+'_next'] = move[uid].shift(-1)
    move = move[move[uid+'_next'] == move[uid]
                ].drop(['stime', 'duration', uid+'_next'], axis=1)
    move = move.rename(columns={'lon': 'slon',
                                'lat': 'slat',
                                'etime': 'stime',
                                'stime_next': 'etime',
                                'LONCOL': 'SLONCOL',
                                'LATCOL': 'SLATCOL',
                                })
    move['duration'] = (
        move['etime'] - move['stime']).dt.total_seconds()
    return stay, move


def mobile_stay_dutation(staydata, col=['stime', 'etime'], start_hour=8, end_hour=20):
    '''
    Input the stay point data to identify the duration during night and day time.

    Parameters
    ----------------
    staydata : DataFrame
        Stay data
    col : List
        The column name, in the order of ['starttime','endtime']
    start_hour : Number
        Start hour of day time
    end_hour : Number
        End hour of day time

    Returns
    ----------------
    duration_night : Series
        Duration at night time
    duration_day : Series
        Duration at day time
    '''

    if (start_hour > end_hour) | (start_hour < 0) | (start_hour > 24) | (end_hour < 0) | (end_hour > 24):
        raise ValueError(
            'end_hour or start_hour error, it should be: 0 <= start_hour <= end_hour <= 24')

    night_hour = start_hour+24-end_hour
    day_hour = end_hour-start_hour
    stay = staydata.copy()
    stime, etime = col
    stay[stime] = pd.to_datetime(stay[stime])
    stay[etime] = pd.to_datetime(stay[etime])

    stay['preday'] = pd.to_datetime(
        stay[stime].dt.date)-pd.Timedelta(1, unit='days')+pd.Timedelta(end_hour, unit='hours')

    tmp = (stay[stime]-stay['preday']).dt.total_seconds()
    days = tmp//(24*3600)
    remain = tmp % (24*3600)
    duration_night_stime = days*night_hour*3600 + \
        np.min([remain, np.ones(len(tmp))*night_hour*3600], axis=0)
    duration_day_stime = days*day_hour*3600 + \
        np.max([remain-(night_hour*3600), np.zeros(len(tmp))], axis=0)

    tmp = (stay[etime]-stay['preday']).dt.total_seconds()
    days = tmp//(24*3600)
    remain = tmp % (24*3600)
    duration_night_etime = days*night_hour*3600 + \
        np.min([remain, np.ones(len(tmp))*night_hour*3600], axis=0)
    duration_day_etime = days*day_hour*3600 + \
        np.max([remain-(night_hour*3600), np.zeros(len(tmp))], axis=0)

    duration_night = duration_night_etime-duration_night_stime
    duration_day = duration_day_etime-duration_day_stime

    return duration_night, duration_day


def mobile_identify_home(staydata, col=['uid','stime', 'etime','LONCOL', 'LATCOL'], start_hour=8, end_hour=20 ):
    '''
    Identify home location from mobile phone stay data. The rule is to identify the locations with longest 
    duration in night time. 

    Parameters
    ----------------
    staydata : DataFrame
        Stay data
    col : List
        The column name, in the order of ['uid','stime', 'etime', 'locationtag1', 'locationtag2', ...].
        There can be multiple 'locationtag' columns to specify the location.
    start_hour, end_hour : Number
        Start hour and end hour of day time

    Returns
    ----------------
    home : DataFrame
        Home location of mobile phone users
    '''
    uid = col[0]
    stime = col[1]
    etime = col[2]
    stay = staydata.copy()
    if ('duration_night' not in stay.columns) | ('duration_day' not in stay.columns):
        stay['duration_night'], stay['duration_day'] = mobile_stay_dutation(
            stay, col = [stime,etime],start_hour=start_hour, end_hour=end_hour)
    # 夜晚最常停留地
    home = stay.groupby([col[0],*col[3:]])['duration_night'].sum().reset_index()
    home = home.sort_values(by=[uid, 'duration_night'], ascending=False).drop_duplicates(
        subset=[uid], keep='first')
    home = home.drop(['duration_night'], axis=1)
    return home


def mobile_identify_work(staydata, col=['uid', 'stime', 'etime', 'LONCOL', 'LATCOL'], minhour=3, start_hour=8, end_hour=20,workdaystart=0, workdayend=4):
    '''
    Identify work location from mobile phone stay data. The rule is to identify the locations with longest 
    duration in day time on weekdays(Average duration should over `minhour`). 

    Parameters
    ----------------
    staydata : DataFrame
        Stay data
    col : List
        The column name, in the order of ['uid','stime','etime', 'locationtag1', 'locationtag2', ...].
        There can be multiple 'locationtag' columns to specify the location.
    minhour : Number
        Minimun duration in work days (hours).
    workdaystart,workdayend : Number
        Start and end of work days in the week. 0 - Monday, 4 - Friday
    start_hour, end_hour : Number
        Start hour and end hour of day time


    Returns
    ----------------
    work : DataFrame
        work location of mobile phone users
    '''
    uid = col[0]
    stime = col[1]
    etime = col[2]
    stay = staydata.copy()

    stay[stime] = pd.to_datetime(stay[stime])
    stay[etime] = pd.to_datetime(stay[etime])

    # 在工作日出现的日期数
    stay_workdays = stay[(stay[stime].dt.weekday >= workdaystart) &
                         (stay[stime].dt.weekday <= workdayend)].copy()

    # 将跨日的活动缩短为当日，以便计算日均持续时间
    stay_workdays['nextday'] = pd.to_datetime(
        stay_workdays[stime].dt.date+pd.Timedelta(1, unit='days'))
    stay_workdays[etime] = stay_workdays[[etime, 'nextday']].min(axis=1)
    stay_workdays['duration_night'], stay_workdays['duration_day'] = mobile_stay_dutation(
        stay_workdays, col = [stime,etime],start_hour=start_hour, end_hour=end_hour)

    # 白天最常活动地
    work = stay_workdays.groupby(
        [col[0], *col[3:]])['duration_day'].sum().reset_index()
    work = work.sort_values(by=[uid, 'duration_day'], ascending=False).drop_duplicates(
        subset=[uid], keep='first')

    # 人出现在多少个工作日
    stay_workdays['date'] = stay_workdays[stime].dt.date
    uid_workdays = stay_workdays[[uid, 'date']].drop_duplicates().groupby([uid])[
        'date'].count().reset_index()

    # 要求工作日每天平均minhour小时以上
    work = pd.merge(work, uid_workdays)
    work = work[(work['duration_day']/work['date']) >= minhour*3600].copy()
    work = work.drop(['duration_day', 'date'], axis=1)

    return work


def mobile_plot_activity(data, col=['stime', 'etime', 'LONCOL', 'LATCOL'],
                         figsize=(10, 5), dpi=250):
    '''
    Plot the activity plot of individual

    Parameters
    ----------------
    data : DataFrame
        activity information of one person
    col : List
        The column name.[starttime,endtime,LONCOL,LATCOL] of activities
    '''
    stime, etime, LONCOL, LATCOL = col
    activity = data.copy()
    activity['date'] = activity[stime].dt.date
    dates = list(activity['date'].astype(str).drop_duplicates())
    dates_all = []
    minday = min(dates)
    maxday = max(dates)
    import datetime
    thisdate = minday
    while thisdate != maxday:
        dates_all.append(thisdate)
        thisdate = str((pd.to_datetime(thisdate+' 00:00:00') +
                       datetime.timedelta(days=1)).date())
    dates = dates_all
    import matplotlib.pyplot as plt
    import numpy as np
    activity['duration'] = (activity[etime]-activity[stime]).dt.total_seconds()
    activity = activity[-activity['duration'].isnull()]
    import time
    activity['ststmp'] = activity[stime].astype(str).apply(
        lambda x: time.mktime(
            time.strptime(x, '%Y-%m-%d %H:%M:%S'))).astype('int64')
    activity['etstmp'] = activity[etime].astype(str).apply(
        lambda x: time.mktime(
            time.strptime(x, '%Y-%m-%d %H:%M:%S'))).astype('int64')
    activityinfo = activity[[LONCOL, LATCOL]].drop_duplicates()
    indexs = list(range(1, len(activityinfo)+1))
    np.random.shuffle(indexs)
    activityinfo['index'] = indexs
    import matplotlib as mpl
    norm = mpl.colors.Normalize(vmin=0, vmax=len(activityinfo))
    from matplotlib.colors import ListedColormap
    import seaborn as sns
    cmap = ListedColormap(sns.hls_palette(
        n_colors=len(activityinfo), l=.5, s=0.8))
    plt.figure(1, figsize, dpi)
    ax = plt.subplot(111)
    plt.sca(ax)
    for day in range(len(dates)):
        plt.bar(day, height=24*3600, bottom=0, width=0.4, color=(0, 0, 0, 0.1))
        stime = dates[day]+' 00:00:00'
        etime = dates[day]+' 23:59:59'
        bars = activity[(activity['stime'] < etime) &
                        (activity['etime'] > stime)].copy()
        bars['ststmp'] = bars['ststmp'] - \
            time.mktime(time.strptime(stime, '%Y-%m-%d %H:%M:%S'))
        bars['etstmp'] = bars['etstmp'] - \
            time.mktime(time.strptime(stime, '%Y-%m-%d %H:%M:%S'))
        for row in range(len(bars)):
            plt.bar(day,
                    height=bars['etstmp'].iloc[row]-bars['ststmp'].iloc[row],
                    bottom=bars['ststmp'].iloc[row],
                    color=cmap(
                        norm(
                            activityinfo[
                                (activityinfo[LONCOL] == bars[LONCOL].
                                 iloc[row]) &
                                (activityinfo[LATCOL] ==
                                 bars[LATCOL].iloc[row])
                            ]['index'].iloc[0])))
    plt.xlim(-0.5, len(dates))
    plt.ylim(0, 24*3600)
    plt.xticks(range(len(dates)), [i[-5:] for i in dates])
    plt.yticks(range(0, 24*3600+1, 3600),
               pd.DataFrame({'t': range(0, 25)})['t'].astype('str')+':00')
    plt.show()

'''Old namespace'''
traj_stay_move = mobile_stay_move
plot_activity = mobile_plot_activity



'''
BSD 3-Clause License

Copyright (c) 2021, Qing Yu
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its
   contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
'''

import pandas as pd
import numpy as np
import geopandas as gpd
from .traj import points_to_traj
from .grids import (
    area_to_params,
    GPS_to_grid,
    grid_to_polygon,
    grid_to_centre,
)
from .odprocess import (
    odagg_grid
)


def visualization_trip(trajdata, col=['Lng', 'Lat', 'ID', 'Time'],
                       zoom='auto', height=500):
    '''
    The input is the trajectory data and the column name. The output is the
    visualization result based on kepler

    Parameters
    -------
    trajdata : DataFrame
        Trajectory points data
    col : List
        The column name, in the sequence of [longitude, latitude, vehicle id,
        time]
    zoom : number
        Map zoom level
    height : number
        The height of the map frame

    Returns
    -------
    vmap : keplergl.keplergl.KeplerGl
        Visualizations provided by keplergl
    '''
    try:
        from keplergl import KeplerGl
    except ImportError:
        raise ImportError(
            "Please install keplergl, run "
            "the following code in cmd: pip install keplergl")
    print('Processing trajectory data...')
    [Lng, Lat, ID, timecol] = col
        #clean data
    trajdata = trajdata[-((trajdata[Lng].isnull())|(trajdata[Lat].isnull()))]
    trajdata = trajdata[(trajdata[Lng]>=-180)&(trajdata[Lng]<=180)&(trajdata[Lat]>=-90)&(trajdata[Lat]<=90)]
    
    trajdata[timecol] = pd.to_datetime(trajdata[timecol])

    trajdata = trajdata.sort_values(by=[ID, timecol])
    traj = points_to_traj(trajdata, col=[Lng, Lat, ID], timecol=timecol)
    ls = []
    for i in range(len(traj['features'])):
        ls.append(traj['features'][i]['geometry']['coordinates'][0])
        ls.append(traj['features'][i]['geometry']['coordinates'][-1])
    ls = pd.DataFrame(ls)
    lon_center, lat_center, starttime = ls[0].mean(), ls[1].mean(), ls[3].min()
    if zoom == 'auto':
        lon_min, lon_max = ls[0].quantile(0.05), ls[0].quantile(0.95)
        zoom = 8.5-np.log(lon_max-lon_min)/np.log(2)
    print('Generate visualization...')
    vmap = KeplerGl(config={
        "version": "v1",
        "config":
        {"visState":
         {
             "filters": [],
             "layers": [
                 {
                     "id": "hizm36i",
                     "type": "trip",
                             "config":
                             {
                                 "dataId": "trajectory",
                                 "label": "trajectory",
                                 "color": [255, 255, 255],
                                 "highlightColor": [255, 255, 0, 255],
                                 "columns":
                                 {
                                     "geojson": "_geojson"
                                 },
                                 "isVisible": True,
                             },
                 }],
             "layerBlending": "additive",
             "animationConfig":
             {
                 "currentTime": starttime,
                 "speed": 0.1
             }
         },
            "mapState":
            {
                "bearing": 0,
                "latitude": lat_center,
                "longitude": lon_center,
                "pitch": 0,
                "zoom": zoom,
         },
         }},
        data={'trajectory': traj}, height=height)
    return vmap


def visualization_od(oddata, col=['slon', 'slat', 'elon', 'elat'],
                     zoom='auto', height=500, accuracy=500, mincount=0):
    '''
    The input is the OD data and the column. The output is the visualization
    result based on kepler

    Parameters
    -------
    oddata : DataFrame
        OD data
    col : List
        The column name. The user can choose a non-weight Origin-Destination
        (OD) data, in the sequence of [origin longitude, origin latitude,
        destination longitude, destination latitude]. For this, The aggregation
        is automatic. Or, the user can also input a weighted OD data, in the
        sequence of [origin longitude, origin latitude, destination longitude,
        destination latitude, count]
    zoom : number
        Map zoom level (Optional). Default value: auto
    height : number
        The height of the map frame
    accuracy : number
        Grid size
    mincount : number
        The minimum OD counts, OD with less counts will not be displayed

    Returns
    -------
    vmap : keplergl.keplergl.KeplerGl
        Visualizations provided by keplergl
    '''
    try:
        from keplergl import KeplerGl
    except ImportError:
        raise ImportError(
            "Please install keplergl, run "
            "the following code in cmd: pip install keplergl")
    import numpy as np
    if len(col) == 4:
        slon, slat, elon, elat = col
        lon1 = oddata[slon].astype(float).quantile(0.01)
        lon2 = oddata[slon].astype(float).quantile(0.99)
        lat1 = oddata[slat].astype(float).quantile(0.01)
        lat2 = oddata[slat].astype(float).quantile(0.99)
        bounds = [lon1, lat1, lon2, lat2]
        params = area_to_params(bounds, accuracy=accuracy)
        od_gdf = odagg_grid(oddata, params, col=col)
        if zoom == 'auto':
            zoom = 8.5-np.log(lon2-lon1)/np.log(2)
        lon_center, lat_center = (lon2+lon1)/2, (lat2+lat1)/2
        lon1, lat1, lon2, lat2, count = (
            'SHBLON',
            'SHBLAT',
            'EHBLON',
            'EHBLAT',
            'count'
        )
    if len(col) == 5:
        slon, slat, elon, elat, count = col
        lon1 = oddata[slon].astype(float).quantile(0.01)
        lon2 = oddata[slon].astype(float).quantile(0.99)
        lat1 = oddata[slat].astype(float).quantile(0.01)
        lat2 = oddata[slat].astype(float).quantile(0.99)
        lon_center, lat_center = (lon2+lon1)/2, (lat2+lat1)/2
        if zoom == 'auto':
            zoom = 8.5-np.log(lon2-lon1)/np.log(2)
        od_gdf = oddata
        lon1, lat1, lon2, lat2, count = col
    od_gdf = od_gdf[od_gdf[count] >= mincount]
    height = 500
    vmap = KeplerGl(config={
        'version': 'v1',
        'config':
        {
            'visState':
            {
                'filters': [],
                'layers': [
                    {
                        'id': 'd3s4dcp',
                        'type': 'arc',
                        'config':
                        {
                            'dataId': 'od',
                            'label': 'shb -> ehb arc',
                            'color': [146, 38, 198],
                            'highlightColor': [252, 242, 26, 255],
                            'columns':
                            {
                                'lat0': lat1,
                                'lng0': lon1,
                                'lat1': lat2,
                                'lng1': lon2
                            },
                            'isVisible': True,
                            'visConfig':
                            {
                                'opacity': 1,
                                'thickness': 2,
                                'colorRange':
                                {
                                    'name': 'Global Warming 8',
                                    'type': 'sequential',
                                    'category': 'Uber',
                                    'colors': ['#4C0035',
                                               '#650031',
                                               '#7F0023',
                                               '#98000A',
                                               '#B21800',
                                               '#CB4600',
                                               '#E57F00',
                                               '#FFC300'
                                               ]
                                },
                                'sizeRange': [0, 3.5],
                                'targetColor': None
                            },
                            'hidden': False,
                            'textLabel': [
                                {
                                    'field': None,
                                    'color': [255, 255, 255],
                                    'size': 18,
                                    'offset': [0, 0],
                                    'anchor': 'start',
                                    'alignment': 'center'
                                }]
                        },
                        'visualChannels':
                        {
                            'colorField':
                            {
                                'name': count,
                                'type': 'integer'
                            },
                            'colorScale': 'quantile',
                            'sizeField':
                            {
                                'name': count,
                                'type': 'integer'
                            },
                            'sizeScale': 'log'
                        }
                    }],
                'interactionConfig':
                {
                    'tooltip':
                    {
                        'fieldsToShow':
                        {
                            'od': [
                                {
                                    'name': lon1,
                                    'format': None
                                },
                                {
                                    'name': lat1,
                                    'format': None
                                },
                                {
                                    'name': lon2,
                                    'format': None
                                },
                                {
                                    'name': lat2,
                                    'format': None
                                },
                                {
                                    'name': count,
                                    'format': None
                                }]
                        },
                        'compareMode': False,
                        'compareType': 'absolute',
                        'enabled': True
                    },
                    'brush':
                    {
                        'size': 0.5,
                        'enabled': False
                    },
                    'geocoder':
                    {
                        'enabled': False
                    },
                    'coordinate':
                    {
                        'enabled': False
                    }
                },
                'layerBlending': 'normal',
                'splitMaps': [],
                'animationConfig':
                {
                    'currentTime': None,
                    'speed': 1
                }
            },
            'mapState':
            {
                'bearing': 24.18348623853211,
                'dragRotate': True,
                'latitude': lat_center,
                'longitude': lon_center,
                'pitch': 23.707784107832463,
                'zoom': zoom,
                'isSplit': False
            },
            'mapStyle':
            {
                'styleType': 'dark',
                'topLayerGroups':
                {},
                'visibleLayerGroups':
                {
                    'label': True,
                    'road': True,
                    'border': False,
                    'building': True,
                    'water': True,
                    'land': True,
                    '3d building': False
                },
                'threeDBuildingColor': [9.665468314072013,
                                        17.18305478057247,
                                        31.1442867897876
                                        ],
                'mapStyles':
                {}
            }
        }
    }, data={'od': od_gdf.to_json()}, height=height)
    return vmap


def visualization_data(data, col=['lon', 'lat'], accuracy=500, height=500,
                       maptype='point', zoom='auto'):
    '''
    The input is the data points, this function will aggregate and then
    visualize it

    Parameters
    -------
    data : DataFrame
        The data point
    col : List
        The column name. The user can choose a non-weight Origin-Destination
        (OD) data, in the sequence of [longitude, latitude]. For this, The
        aggregation is automatic. Or, the user can also input a weighted OD
        data, in the sequence of [longitude, latitude, count]
    zoom : number
        Map zoom level (Optional). Default value: auto
    height : number
        The height of the map frame
    accuracy : number
        Grid size
    maptype : str
        Map type, ‘point’ or ‘heatmap’

    Returns
    -------
    vmap : keplergl.keplergl.KeplerGl
        Visualizations provided by keplergl
    '''
    try:
        from keplergl import KeplerGl
    except ImportError:
        raise ImportError(
            "Please install keplergl, run "
            "the following code in cmd: pip install keplergl")

    if len(col) == 2:
        lon, lat = col[0], col[1]
        count = 'count'

        data[lon] = data[lon].astype('float')
        data[lat] = data[lat].astype('float')
        #clean data
        data = data[-((data[lon].isnull())|(data[lat].isnull()))]
        data = data[(data[lon]>=-180)&(data[lon]<=180)&(data[lat]>=-90)&(data[lat]<=90)]

        bounds = [data[lon].min(), data[lat].min(),
                  data[lon].max(), data[lat].max()]
        lon_center, lat_center = data[lon].mean(), data[lat].mean()
        if zoom == 'auto':
            lon_min, lon_max = data[lon].quantile(
                0.05), data[lon].quantile(0.95)
            zoom = 8.5-np.log(lon_max-lon_min)/np.log(2)
        params = area_to_params(bounds, accuracy=accuracy)
        data['LONCOL'], data['LATCOL'] = GPS_to_grid(
            data[lon], data[lat], params)
        data[count] = 1
        data = data.groupby(['LONCOL', 'LATCOL'])[
            'count'].sum().reset_index().reset_index()
        data['geometry'] = grid_to_polygon(
            [data['LONCOL'], data['LATCOL']], params)
        data[lon], data[lat] = grid_to_centre(
            [data['LONCOL'], data['LATCOL']], params)
        data = gpd.GeoDataFrame(data)

    if len(col) == 3:
        lon, lat, count = col
        
        data[lon] = data[lon].astype('float')
        data[lat] = data[lat].astype('float')
        #clean data
        data = data[-((data[lon].isnull())|(data[lat].isnull()))]
        data = data[(data[lon]>=-180)&(data[lon]<=180)&(data[lat]>=-90)&(data[lat]<=90)]

        bounds = [data[lon].min(), data[lat].min(),
                  data[lon].max(), data[lat].max()]
        lon_center, lat_center = data[lon].mean(), data[lat].mean()
        if zoom == 'auto':
            lon_min, lon_max = data[lon].quantile(
                0.05), data[lon].quantile(0.95)
            zoom = 8.5-np.log(lon_max-lon_min)/np.log(2)
        params = area_to_params(bounds, accuracy=accuracy)
        data['LONCOL'], data['LATCOL'] = GPS_to_grid(
            data[lon], data[lat], params)
        data = data.groupby(['LONCOL', 'LATCOL'])[count].sum().reset_index()
        data['geometry'] = grid_to_polygon(
            [data['LONCOL'], data['LATCOL']], params)
        data[lon], data[lat] = grid_to_centre(
            [data['LONCOL'], data['LATCOL']], params)

        data = gpd.GeoDataFrame(data)

    if maptype == 'heatmap':
        vmap = KeplerGl(config={
            'version': 'v1',
            'config': {
                'visState': {
                    'filters': [],
                    'layers': [
                        {'id': 'vpefba0o',
                         'type': 'heatmap',
                         'config': {
                             'dataId': 'data',
                             'label': 'Point',
                             'color': [18, 147, 154],
                             'highlightColor': [252, 242, 26, 255],
                             'columns': {'lat': lat, 'lng': lon},
                             'isVisible': True,
                             'visConfig': {
                                 'opacity': 0.8,
                                 'colorRange': {
                                     'name': 'Global Warming',
                                     'type': 'sequential',
                                     'category': 'Uber',
                                     'colors': ['#5A1846',
                                                '#900C3F',
                                                '#C70039',
                                                '#E3611C',
                                                '#F1920E',
                                                '#FFC300']},
                                 'radius': 23},
                             'hidden': False,
                             'textLabel': [{'field': None,
                                            'color': [255, 255, 255],
                                            'size': 18,
                                            'offset': [0, 0],
                                            'anchor': 'start',
                                            'alignment': 'center'}]},
                         'visualChannels': {
                             'weightField': {'name': count, 'type': 'integer'},
                             'weightScale': 'linear'}}],
                    'interactionConfig': {
                        'tooltip': {
                            'fieldsToShow': {
                                'data': [{
                                    'name': count,
                                    'format': None}]},
                            'compareMode': False,
                            'compareType': 'absolute',
                            'enabled': True},
                        'brush': {'size': 0.5, 'enabled': False},
                        'geocoder': {'enabled': False},
                        'coordinate': {'enabled': False}},
                    'layerBlending': 'normal',
                    'splitMaps': [],
                    'animationConfig': {'currentTime': None, 'speed': 1}},
                'mapState':
                {
                    'bearing': 0,
                    'dragRotate': True,
                    'latitude': lat_center,
                    'longitude': lon_center,
                    'pitch': 0,
                    'zoom': zoom,
                    'isSplit': False
                },
                'mapStyle':
                {
                    'styleType': 'dark',
                    'topLayerGroups':
                    {},
                    'visibleLayerGroups':
                    {
                        'label': True,
                        'road': True,
                        'border': False,
                        'building': True,
                        'water': True,
                        'land': True,
                        '3d building': False
                    },
                    'threeDBuildingColor': [9.665468314072013,
                                            17.18305478057247,
                                            31.1442867897876
                                            ],
                    'mapStyles':
                    {}
                }}}, data={'data': data.to_json()}, height=height)
    else:
        vmap = KeplerGl(config={
            'version': 'v1',
            'config': {
                'visState': {
                    'filters': [],
                    'layers': [{
                        'id': 'ytak0zp',
                        'type': 'geojson',
                        'config': {
                            'dataId': count,
                            'label': count,
                            'color': [77, 193, 156],
                            'highlightColor': [252, 242, 26, 255],
                            'columns': {'geojson': '_geojson'},
                            'isVisible': True,
                            'visConfig': {
                                'opacity': 0.8,
                                'strokeOpacity': 0.8,
                                'thickness': 0.5,
                                'strokeColor': [218, 112, 191],
                                'colorRange': {
                                    'name': 'Global Warming',
                                    'type': 'sequential',
                                    'category': 'Uber',
                                    'colors': ['#5A1846',
                                               '#900C3F',
                                               '#C70039',
                                               '#E3611C',
                                               '#F1920E',
                                               '#FFC300']},
                                'strokeColorRange': {'name': 'Global Warming',
                                                     'type': 'sequential',
                                                     'category': 'Uber',
                                                     'colors': ['#5A1846',
                                                                '#900C3F',
                                                                '#C70039',
                                                                '#E3611C',
                                                                '#F1920E',
                                                                '#FFC300']},
                                'radius': 10,
                                'sizeRange': [0, 10],
                                'radiusRange': [0, 50],
                                'heightRange': [0, 500],
                                'elevationScale': 5,
                                'enableElevationZoomFactor': True,
                                'stroked': False,
                                'filled': True,
                                'enable3d': False,
                                'wireframe': False},
                            'hidden': False,
                            'textLabel': [{'field': None,
                                           'color': [255, 255, 255],
                                           'size': 18,
                                           'offset': [0, 0],
                                           'anchor': 'start',
                                           'alignment': 'center'}]},
                        'visualChannels': {
                            'colorField': {'name': count, 'type': 'integer'},
                            'colorScale': 'quantile',
                            'strokeColorField': None,
                            'strokeColorScale': 'quantile',
                            'sizeField': None,
                            'sizeScale': 'linear',
                            'heightField': None,
                            'heightScale': 'linear',
                            'radiusField': None,
                            'radiusScale': 'linear'}}],
                    'layerBlending': 'normal',
                    'splitMaps': [],
                    'animationConfig': {'currentTime': None, 'speed': 1}},
                'mapState': {'bearing': 0,
                             'dragRotate': False,
                             'latitude': data[lat].mean(),
                             'longitude': data[lon].mean(),
                             'pitch': 0,
                             'zoom': 10,
                             'isSplit': False}}},
            data={count: data.to_json()}, height=height)

    return vmap



'''
BSD 3-Clause License

Copyright (c) 2021, Qing Yu
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its
   contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
'''

import geopandas as gpd
import pandas as pd
import numpy as np
from shapely.geometry import LineString
import shapely
from .preprocess import (
    clean_same,
    clean_outofshape,
    id_reindex
)


def busgps_arriveinfo(data, line, stop, col=[
        'VehicleId', 'GPSDateTime', 'lon', 'lat', 'stopname'],
        stopbuffer=200, mintime=300, project_epsg=2416,
        timegap=1800, method='project', projectoutput=False):
    '''
    Input bus GPS data, bus route and station GeoDataFrame, this
    method can identify the bus arrival and departure information

    Parameters
    -------
    data : DataFrame
        Bus GPS data. It should be the data from one bus route,
        and need to contain vehicle ID, GPS time, latitude and
        longitude (wgs84)
    line : GeoDataFrame
        GeoDataFrame for the bus line
    stop : GeoDataFrame
        GeoDataFrame for bus stops
    col : List
        Column names, in the order of [vehicle ID, time, longitude,
        latitude, station name]
    stopbuffer : number
        Meter. When the vehicle approaches the station within this
        certain distance, it is considered to be arrive at the station.
    mintime : number
        Seconds. Within a short period of time that the bus arrive
        at bus station again, it will not be consider as another arrival
    project_epsg : number
        The matching algorithm will convert the data into a projection
        coordinate system to calculate the distance, here the epsg code
        of the projection coordinate system is given
    timegap : number
        Seconds. For how long the vehicle does not appear, it will be
        considered as a new vehicle
    method : str
        The method of matching the bus, either ‘project’ or ‘dislimit’;
        ‘project’ is to directly match the nearest point on the route,
        which is fast. ‘dislimit’ needs to consider the front point position
        with the distance limitation, the matching speed is slow.
    projectoutput : bool
        Whether to output the projected data

    Returns
    -------
    arrive_info : DataFrame
        Bus arrival and departure information
    '''
    VehicleId, GPSDateTime, lon, lat, stopcol = col
    # Clean data
    print('Cleaning data', end='')
    line.set_crs(crs='epsg:4326', allow_override=True, inplace=True)
    line = line.to_crs(epsg=project_epsg)
    line_buffer = line.copy()
    line_buffer['geometry'] = line_buffer.buffer(200)
    line_buffer = line_buffer.to_crs(epsg=4326)
    print('.', end='')
    data = clean_same(data, col=[VehicleId, GPSDateTime, lon, lat])
    print('.', end='')
    data = clean_outofshape(data, line_buffer, col=[lon, lat], accuracy=500)
    print('.')
    data = id_reindex(data, VehicleId, timegap=timegap,
                      timecol=GPSDateTime, suffix='')

    print('Position matching', end='')
    # project data points onto bus LineString
    lineshp = line['geometry'].iloc[0]
    print('.', end='')
    data['geometry'] = gpd.points_from_xy(data[lon], data[lat])
    data = gpd.GeoDataFrame(data)
    data.set_crs(crs='epsg:4326', allow_override=True, inplace=True)
    print('.', end='')
    data = data.to_crs(epsg=project_epsg)
    print('.', end='')
    if method == 'project':
        data['project'] = data['geometry'].apply(lineshp.project)
    elif method == 'dislimit':
        tmps = []
        # Distance limit method
        for vid in data[VehicleId].drop_duplicates():
            print('.', end='')
            tmp = data[data[VehicleId] == vid].copy()
            i = 0
            tmp = tmp.sort_values(
                by=[VehicleId, GPSDateTime]).reset_index(drop=True)
            tmp['project'] = 0
            for i in range(len(tmp)-1):
                if i == 0:
                    proj = lineshp.project(tmp.iloc[i]['geometry'])
                    tmp.loc[i, 'project'] = proj
                else:
                    proj = tmp['project'].iloc[i]
                dis = tmp.iloc[i +
                               1]['geometry'].distance(tmp.iloc[i]['geometry'])
                if dis == 0:
                    proj1 = proj
                else:
                    proj2 = lineshp.project(tmp.iloc[i+1]['geometry'])
                    if abs(proj2-proj) > dis:
                        proj1 = np.sign(proj2-proj)*dis+proj
                    else:
                        proj1 = proj2
                tmp.loc[i+1, 'project'] = proj1
            tmps.append(tmp)
        data = pd.concat(tmps)
    print('.', end='')
    # Project bus stop to bus line
    stop = stop.to_crs(epsg=project_epsg)
    stop['project'] = stop['geometry'].apply(lineshp.project)
    print('.', end='')
    starttime = data[GPSDateTime].min()
    data['time_st'] = (data[GPSDateTime]-starttime).dt.total_seconds()
    BUS_project = data
    print('.')

    ls = []
    print('Matching arrival and leaving info...', end='')
    for car in BUS_project[VehicleId].drop_duplicates():
        print('.', end='')
        # Extract bus trajectory
        tmp = BUS_project[BUS_project[VehicleId] == car]
        if len(tmp) > 1:
            for stopname in stop[stopcol].drop_duplicates():
                # Get the stop position
                position = stop[stop[stopcol] == stopname]['project'].iloc[0]
                # Identify arrival and departure by intersection ofstop buffer
                # and line segment
                buffer_polygon = LineString([[0, position],
                                             [data['time_st'].max(), position]
                                             ]).buffer(stopbuffer)
                bus_linestring = LineString(tmp[['time_st', 'project']].values)
                line_intersection = bus_linestring.intersection(buffer_polygon)
                # Extract leave time
                if line_intersection.is_empty:
                    # If empty, no bus arrive
                    continue
                else:
                    if type(
                        line_intersection
                    ) == shapely.geometry.linestring.LineString:
                        arrive = [line_intersection]
                    else:
                        arrive = list(line_intersection)
                arrive = pd.DataFrame(arrive)
                arrive['arrivetime'] = arrive[0].apply(
                    lambda r: r.coords[0][0])
                arrive['leavetime'] = arrive[0].apply(
                    lambda r: r.coords[-1][0])
                # Filtering arrival information through time threshold
                a = arrive[['arrivetime']].copy()
                a.columns = ['time']
                a['flag'] = 1
                b = arrive[['leavetime']].copy()
                b.columns = ['time']
                b['flag'] = 0
                c = pd.concat([a, b]).sort_values(by='time')
                c['time1'] = c['time'].shift(-1)
                c['flag_1'] = ((c['time1']-c['time']) <
                               mintime) & (c['flag'] == 0)
                c['flag_2'] = c['flag_1'].shift().fillna(False)
                c['flag_3'] = c['flag_1'] | c['flag_2']
                c = c[-c['flag_3']]
                arrive_new = c[c['flag'] == 1][['time']].copy()
                arrive_new.columns = ['arrivetime']
                arrive_new['leavetime'] = list(c[c['flag'] == 0]['time'])
                arrive_new[stopcol] = stopname
                arrive_new[VehicleId] = car
                # Save data
                ls.append(arrive_new)
    # Concat data
    arrive_info = pd.concat(ls)
    arrive_info['arrivetime'] = starttime + \
        arrive_info['arrivetime'].apply(
            lambda r: pd.Timedelta(int(r), unit='s'))
    arrive_info['leavetime'] = starttime + \
        arrive_info['leavetime'].apply(
            lambda r: pd.Timedelta(int(r), unit='s'))
    if projectoutput:
        return arrive_info, data
    else:
        return arrive_info


def busgps_onewaytime(arrive_info, start, end,
                      col=['VehicleId', 'stopname',
                           'arrivetime', 'leavetime']):
    '''
    Input the departure information table drive_info and the station
    information table stop to calculate the one-way travel time

    Parameters
    -------
    arrive_info : DataFrame
        The departure information table drive_info
    start : Str
        Starting station name
    end : Str
        Ending station name
    col : List
        Column name [vehicle ID, station name,arrivetime,leavetime]


    Returns
    -------
    onewaytime : DataFrame
        One-way travel time of the bus
    '''
    # For one direction
    # The information of start and end points is extracted and merged together
    # Arrival time of terminal
    [VehicleId, stopname, arrivetime, leavetime] = col
    arrive_info[arrivetime] = pd.to_datetime(arrive_info[arrivetime])
    arrive_info[leavetime] = pd.to_datetime(arrive_info[leavetime])
    a = arrive_info[arrive_info[stopname] ==
                    end][[arrivetime, stopname, VehicleId]]
    # Departure time of starting station
    b = arrive_info[arrive_info[stopname] ==
                    start][[leavetime, stopname, VehicleId]]
    a.columns = ['time', stopname, VehicleId]
    b.columns = ['time', stopname, VehicleId]
    # Concat data
    c = pd.concat([a, b])
    # After sorting, extract the travel time of each one-way trip
    c = c.sort_values(by=[VehicleId, 'time'])
    for i in c.columns:
        c[i+'1'] = c[i].shift(-1)
    c = c[(c[VehicleId] == c[VehicleId+'1']) &
          (c[stopname] == start) &
          (c[stopname+'1'] == end)]
    # Calculate the duration of the trip
    c['duration'] = (c['time1'] - c['time']).dt.total_seconds()
    c['shour'] = c['time'].dt.hour
    c['direction'] = start+'-'+end
    c1 = c.copy()
    # Do the same for the other direction
    a = arrive_info[arrive_info[stopname] ==
                    start][['arrivetime', stopname, VehicleId]]
    b = arrive_info[arrive_info[stopname] ==
                    end][['leavetime', stopname, VehicleId]]
    a.columns = ['time', stopname, VehicleId]
    b.columns = ['time', stopname, VehicleId]
    c = pd.concat([a, b])
    c = c.sort_values(by=[VehicleId, 'time'])
    for i in c.columns:
        c[i+'1'] = c[i].shift(-1)
    c = c[(c[VehicleId] == c[VehicleId+'1']) &
          (c[stopname] == end) & (c[stopname+'1'] == start)]
    c['duration'] = (c['time1'] - c['time']).dt.total_seconds()
    c['shour'] = c['time'].dt.hour
    c['direction'] = end+'-'+start
    c2 = c.copy()
    onewaytime = pd.concat([c1, c2])
    return onewaytime



'''
BSD 3-Clause License

Copyright (c) 2021, Qing Yu
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its
   contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
'''

import geopandas as gpd
import pandas as pd
import numpy as np
from .preprocess import id_reindex


def traj_densify(data, col=['Vehicleid', 'Time', 'Lng', 'Lat'], timegap=15):
    '''
    Trajectory densification, ensure that there is a trajectory point each
    timegap seconds

    Parameters
    -------
    data : DataFrame
        Data
    col : List
        The column name, in the sequence of [Vehicleid, Time, lng, lat]
    timegap : number
        The sampling interval (second)

    Returns
    -------
    data1 : DataFrame
        The processed data
    '''
    Vehicleid, Time, Lng, Lat = col
    data[Time] = pd.to_datetime(data[Time])
    data1 = data.copy()
    data1 = data1.drop_duplicates([Vehicleid, Time])
    data1 = id_reindex(data1, Vehicleid)
    data1 = data1.sort_values(by=[Vehicleid+'_new', Time])
    data1['utctime'] = data1[Time].apply(lambda r: int(r.value/1000000000))
    data1['utctime_new'] = data1[Vehicleid+'_new']*10000000000+data1['utctime']
    a = data1.groupby([Vehicleid+'_new']
                      )['utctime'].min().rename('mintime').reset_index()
    b = data1.groupby([Vehicleid+'_new']
                      )['utctime'].max().rename('maxtime').reset_index()
    minmaxtime = pd.merge(a, b)
    mintime = data1['utctime'].min()
    maxtime = data1['utctime'].max()
    timedata = pd.DataFrame(range(mintime, maxtime, timegap), columns=[Time])
    timedata['tmp'] = 1
    minmaxtime['tmp'] = 1
    minmaxtime = pd.merge(minmaxtime, timedata)
    minmaxtime = minmaxtime[(minmaxtime['mintime'] <= minmaxtime[Time]) & (
        minmaxtime['maxtime'] >= minmaxtime[Time])]
    minmaxtime['utctime_new'] = minmaxtime[Vehicleid+'_new'] * \
        10000000000+minmaxtime[Time]
    minmaxtime[Time] = pd.to_datetime(minmaxtime[Time], unit='s')
    data1 = pd.concat([data1, minmaxtime[['utctime_new', Time]]]
                      ).sort_values(by=['utctime_new'])
    data1 = data1.drop_duplicates(['utctime_new'])
    data1[Lng] = data1.set_index('utctime_new')[
        Lng].interpolate(method='index').values
    data1[Lat] = data1.set_index('utctime_new')[
        Lat].interpolate(method='index').values
    data1[Vehicleid] = data1[Vehicleid].ffill()
    data1[Vehicleid] = data1[Vehicleid].bfill()
    data1 = data1.drop([Vehicleid+'_new', 'utctime', 'utctime_new'], axis=1)
    return data1


def traj_sparsify(data, col=['Vehicleid', 'Time', 'Lng', 'Lat'], timegap=15,
                  method='subsample'):
    '''
    Trajectory sparsify. When the sampling frequency of trajectory data is too
    high, the amount of data is too large, which is not convenient for the
    analysis of some studies that require less data frequency. This function
    can expand the sampling interval and reduce the amount of data.

    Parameters
    -------
    data : DataFrame
        Data
    col : List
        The column name, in the sequence of [Vehicleid, Time, lng, lat]
    timegap : number
        Time gap between trajectory point
    method : str
        'interpolate' or 'subsample'

    Returns
    -------
    data1 : DataFrame
        Sparsified trajectory data
    '''
    Vehicleid, Time, Lng, Lat = col
    data[Time] = pd.to_datetime(data[Time], unit='s')
    data1 = data.copy()
    data1 = data1.drop_duplicates([Vehicleid, Time])
    data1 = id_reindex(data1, Vehicleid)
    data1 = data1.sort_values(by=[Vehicleid+'_new', Time])
    data1['utctime'] = data1[Time].apply(lambda r: int(r.value/1000000000))
    data1['utctime_new'] = data1[Vehicleid+'_new']*10000000000+data1['utctime']
    if method == 'interpolate':
        a = data1.groupby([Vehicleid+'_new']
                          )['utctime'].min().rename('mintime').reset_index()
        b = data1.groupby([Vehicleid+'_new']
                          )['utctime'].max().rename('maxtime').reset_index()
        minmaxtime = pd.merge(a, b)
        mintime = data1['utctime'].min()
        maxtime = data1['utctime'].max()
        timedata = pd.DataFrame(
            range(mintime, maxtime, timegap), columns=[Time])
        timedata['tmp'] = 1
        minmaxtime['tmp'] = 1
        minmaxtime = pd.merge(minmaxtime, timedata)
        minmaxtime = minmaxtime[(minmaxtime['mintime'] <= minmaxtime[Time]) & (
            minmaxtime['maxtime'] >= minmaxtime[Time])]
        minmaxtime['utctime_new'] = minmaxtime[Vehicleid+'_new'] * \
            10000000000+minmaxtime[Time]
        minmaxtime[Time] = pd.to_datetime(minmaxtime[Time], unit='s')
        data1 = pd.concat([
            data1, minmaxtime[['utctime_new', Time]]
        ]).sort_values(by=['utctime_new'])
        data1 = data1.drop_duplicates(['utctime_new'])
        data1[Lng] = data1.set_index('utctime_new')[
            Lng].interpolate(method='index').values
        data1[Lat] = data1.set_index('utctime_new')[
            Lat].interpolate(method='index').values
        data1[Vehicleid] = data1[Vehicleid].ffill()
        data1[Vehicleid] = data1[Vehicleid].bfill()
        data1 = pd.merge(minmaxtime['utctime_new'], data1)
        data1 = data1.drop(
            [Vehicleid+'_new', 'utctime', 'utctime_new'], axis=1)
    if method == 'subsample':
        data1['utctime_new'] = (data1['utctime_new']/timegap).astype(int)
        data1 = data1.drop_duplicates(subset=['utctime_new'])
        data1 = data1.drop(
            [Vehicleid+'_new', 'utctime', 'utctime_new'], axis=1)
    return data1


def points_to_traj(traj_points, col=['Lng', 'Lat', 'ID'], timecol=None):
    '''
    Input trajectory, generate GeoDataFrame

    Parameters
    -------
    traj_points : DataFrame
        trajectory data
    col : List
        The column name, in the sequence of [lng, lat,trajectoryid]
    timecol : str(Optional)
        Optional, the column name of the time column. If given, the geojson
        with [longitude, latitude, altitude, time] in returns can be put into
        the Kepler to visualize the trajectory

    Returns
    -------
    traj : GeoDataFrame
        Generated trajectory
    '''
    [Lng, Lat, ID] = col
    if timecol:
        geometry = []
        traj_id = []
        for i in traj_points[ID].drop_duplicates():
            coords = traj_points[traj_points[ID] == i][[Lng, Lat, timecol]]
            coords[timecol] = coords[timecol].apply(
                lambda r: int(r.value/1000000000))
            coords['altitude'] = 0
            coords = coords[[Lng, Lat, 'altitude', timecol]].values.tolist()
            traj_id.append(i)
            if len(coords) >= 2:
                geometry.append({
                    "type": "Feature",
                    "properties": {"ID":  i},
                    "geometry": {"type": "LineString",
                                 "coordinates": coords}})
        traj = {"type": "FeatureCollection",
                "features": geometry}
    else:
        traj = gpd.GeoDataFrame()
        from shapely.geometry import LineString
        geometry = []
        traj_id = []
        for i in traj_points[ID].drop_duplicates():
            coords = traj_points[traj_points[ID] == i][[Lng, Lat]].values
            traj_id.append(i)
            if len(coords) >= 2:
                geometry.append(LineString(coords))
            else:
                geometry.append(None)
        traj[ID] = traj_id
        traj['geometry'] = geometry
        traj = gpd.GeoDataFrame(traj)
    return traj


def dumpjson(data, path):
    '''
    Input the json data and save it as a file. This method is suitable for
    sovling the problem that numpy cannot be compatiable with json package.

    Parameters
    -------
    data : json
        The json data to be saved


    path : str
        The storage path

    '''
    import json

    class NpEncoder(json.JSONEncoder):
        def default(self, obj):
            if isinstance(obj, np.integer):
                return int(obj)
            elif isinstance(obj, np.floating):
                return float(obj)
            elif isinstance(obj, np.ndarray):
                return obj.tolist()
            else:
                return super(NpEncoder, self).default(obj)
    f = open(path, mode='w')
    json.dump(data, f, cls=NpEncoder)
    f.close()



'''
BSD 3-Clause License

Copyright (c) 2021, Qing Yu
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its
   contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
'''

import numpy as np
import math
import urllib
import io
from PIL import Image
import os
import sys
import warnings


def searchfile(filename):
    '''
    Search for mapboxtoken file
    '''
    fileroot = ''
    for i in sys.path:
        try:
            if filename in os.listdir(i):
                fileroot = i
        except Exception:
            pass
    if fileroot == '':
        fileroot = sys.path[-1]
    return fileroot+'/'+filename


def set_mapboxtoken(mapboxtoken):
    '''
    Set mapboxtoken

    Parameters
    -------
    mapboxtoken : str
        mapboxtoken
    '''
    filepath = searchfile('mapboxtoken.txt')
    f = open(filepath, mode='w')
    f.write(mapboxtoken)
    f.close()
    print('Success')


def read_mapboxtoken():
    '''
    Read mapboxtoken
    '''
    filepath = searchfile('mapboxtoken.txt')
    try:
        f = open(filepath, mode='r')
        mapboxtoken = f.readline()
        f.close()
    except Exception:
        warnings.warn("Mapboxtoken not found, "
                      "The basemap is set as OpenStreetMap"
                      "please use tbd.set_mapboxtoken() to set the access token, see: "
                      "https://transbigdata.readthedocs.io/en/latest/plot_map.html"
                      )
        mapboxtoken = ''
    return mapboxtoken


def set_imgsavepath(imgsavepath):
    '''
    Set savepath for maps

    Parameters
    -------
    imgsavepath : str
        savepath
    '''
    filepath = searchfile('imgsavepath.txt')
    f = open(filepath, mode='w')
    f.write(imgsavepath)
    f.close()
    print('Success')


def read_imgsavepath():
    '''
    Read map savepath
    '''
    filepath = searchfile('imgsavepath.txt')
    try:
        f = open(filepath, mode='r')
        imgsavepath = f.readline()
        f.close()
    except:
        warnings.warn('Map base map storage path not found, \
        please use tbd.set_imgsavepath() to set it first, \
        see: https://transbigdata.readthedocs.io/en/latest/plot_map.html')
        imgsavepath=''
    return imgsavepath


def deg2num(lat_deg, lon_deg, zoom):
    '''
    Calculate xy tiles from coordinates

    Parameters
    -------
    lon_deg : number
        Longitude
    lat_deg : number
        Latitude
    zoom : Int
        Zoom level of the map
    '''
    lat_rad = math.radians(lat_deg)
    n = 2.0 ** zoom
    xtile = int((lon_deg + 180.0) / 360.0 * n)
    ytile = int((1.0 - math.log(math.tan(lat_rad) +
                (1 / math.cos(lat_rad))) / math.pi) / 2.0 * n)
    return (xtile, ytile)


def num2deg(xtile, ytile, zoom):
    '''
    Calculate coordinates from xy tiles

    Parameters
    -------
    xtile : Int
        x tile
    ytile : Int
        y tile
    zoom : Int
        Zoom level of the map
    '''
    n = 2.0 ** zoom
    lon_deg = xtile / n * 360.0 - 180.0
    lat_rad = math.atan(math.sinh(math.pi * (1 - 2 * ytile / n)))
    lat_deg = math.degrees(lat_rad)
    return (lat_deg, lon_deg)


def getImageCluster(lon_deg, lat_deg, delta_long, delta_lat, zoom,
                    printlog, imgsavepath, style=4, access_token=''):
    '''
    Get map image
    '''
    smurl = ''
    if (style == 1) | (style == 'streets'):
        styleid = 'ckwinzgw581od14mpyfhka6nk'
        smurl = r'https://api.mapbox.com/styles/v1/ni1o1/'+styleid + \
            r'/tiles/256/{0}/{1}/{2}?&access_token='+access_token
    if (style == 2) | (style == 'outdoors'):
        styleid = 'ckwinx7aj4y4a15p7ftfwq9dn'
        smurl = r'https://api.mapbox.com/styles/v1/ni1o1/'+styleid + \
            r'/tiles/256/{0}/{1}/{2}?&access_token='+access_token
    if (style == 3) | (style == 'satellite'):
        styleid = 'cjv36cj9u4h1q1ftemjed4f2y'
        smurl = r'https://api.mapbox.com/styles/v1/ni1o1/'+styleid + \
            r'/tiles/256/{0}/{1}/{2}?&access_token='+access_token
    if (style == 4) | (style == 'light'):
        styleid = 'ckwfx658z4dpb14ocnz6tky9d'
        smurl = r'https://api.mapbox.com/styles/v1/ni1o1/'+styleid + \
            r'/tiles/256/{0}/{1}/{2}?&access_token='+access_token
    if (style == 5) | (style == 'dark'):
        styleid = 'cjetnd20i1vbi2qqxbh0by7p8'
        smurl = r'https://api.mapbox.com/styles/v1/ni1o1/'+styleid + \
            r'/tiles/256/{0}/{1}/{2}?&access_token='+access_token
    if (style == 6) | (style == 'light-ch'):
        styleid = 'ckj9bhq7s9mvj19mq3e3fye35'
        smurl = r'https://api.mapbox.com/styles/v1/ni1o1/'+styleid + \
            r'/tiles/256/{0}/{1}/{2}?&access_token='+access_token
    if (style == 7) | (style == 'ice creem'):
        styleid = 'cjv36iiz9243t1fo8mweb4z6r'
        smurl = r'https://api.mapbox.com/styles/v1/ni1o1/'+styleid + \
            r'/tiles/256/{0}/{1}/{2}?&access_token='+access_token
    if (style == 8) | (style == 'night'):
        styleid = 'ck2o3fyvy0dch1cp6j2pkz2dv'
        smurl = r'https://api.mapbox.com/styles/v1/ni1o1/'+styleid + \
            r'/tiles/256/{0}/{1}/{2}?&access_token='+access_token
    if (style == 9) | (style == 'terrain'):
        styleid = 'cjv36gyklf43q1fnuwibiuetl'
        smurl = r'https://api.mapbox.com/styles/v1/ni1o1/'+styleid + \
            r'/tiles/256/{0}/{1}/{2}?&access_token='+access_token
    if (style == 10) | (style == 'basic blue'):
        styleid = 'ckwio2ze12fgk15p2alr5a4xj'
        smurl = r'https://api.mapbox.com/styles/v1/ni1o1/'+styleid + \
            r'/tiles/256/{0}/{1}/{2}?&access_token='+access_token
    if (style == 11) :
        styleid = 'cl39hgul1000514llp3yj7yh3'
        smurl = r'https://api.mapbox.com/styles/v1/ni1o1/'+styleid + \
            r'/tiles/256/{0}/{1}/{2}?&access_token='+access_token
    if (style == 12) :
        styleid = 'cl38pljx0006r14qp7ioy7gcc'
        smurl = r'https://api.mapbox.com/styles/v1/ni1o1/'+styleid + \
            r'/tiles/256/{0}/{1}/{2}?&access_token='+access_token
    if (style == 0) | (style == 'OSM'):
        styleid = 'osm'
        smurl = r'https://tile.openstreetmap.org/{0}/{1}/{2}.png'
    if (str(style)[:16] == 'mapbox://styles/'):
        styleid = style.split('/')[-1]
        smurl = r'https://api.mapbox.com/styles/v1/'+style[16:] + \
            r'/tiles/256/{0}/{1}/{2}?&access_token='+access_token
    if smurl == '':
        raise ValueError('Style error')

    xmin, ymax = deg2num(lat_deg, lon_deg, zoom)
    xmax, ymin = deg2num(lat_deg + delta_lat, lon_deg + delta_long, zoom)

    def get_img(smurl, zoom, xtile, ytile, imgsize, imgsavepath):
        import os

        filename = str(style)+str(styleid)+'-'+str(zoom)+'-' + \
            str(xtile)+'-'+str(ytile)+'-'+str(imgsize)+'.png'

        def savefig(filename, tile):
            try:
                if 'tileimg' in os.listdir(imgsavepath):
                    if filename in os.listdir(imgsavepath+'tileimg'):
                        pass
                    else:
                        tile.save(imgsavepath+'tileimg/'+filename)
                        if printlog:
                            print('figsaved:'+imgsavepath+'tileimg/'+filename)
                else:
                    os.mkdir(imgsavepath+'tileimg')
            except Exception:
                pass

        def loadfig(filename):
            try:
                if 'tileimg' in os.listdir(imgsavepath):
                    if filename in os.listdir(imgsavepath+'tileimg'):
                        tile = Image.open(imgsavepath+'tileimg\\'+filename)
                        return tile
                    else:
                        return None
                else:
                    os.mkdir(imgsavepath+'tileimg')
                    return None
            except Exception:
                return None

        tile = loadfig(filename)
        if tile is None:
            try:
                t = 0
                while t < 10:
                    try:
                        imgurl = smurl.format(zoom, xtile, ytile)
                        header = {}
                        header['Accept'] = 'text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,*/*;q=0.8'
                        header['Accept-Language'] = 'zh-CN,zh;q=0.8'
                        header['Upgrade-Insecure-Requests'] = '1'
                        header['User-Agent'] = 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/51.0.2704.84 Safari/537.36'
                        request = urllib.request.Request(imgurl)
                        for k, v in header.items():
                            request.add_header(k, v)
                        imgstr = urllib.request.urlopen(
                            request, timeout=6).read()
                        tile = Image.open(io.BytesIO(imgstr))
                        savefig(filename, tile)
                        Cluster.paste(tile, box=(
                            (xtile-xmin)*imgsize,  (ytile-ymin)*imgsize))
                        t = 10
                    except Exception:
                        if printlog:
                            print('Get map tile failed, retry ', t)
                        t += 1
            except Exception:
                print("Couldn't download image")
                tile = None
        else:
            Cluster.paste(tile, box=(
                (xtile-xmin)*imgsize,  (ytile-ymin)*imgsize))

    imgsize = 256
    import threading
    threads = []
    Cluster = Image.new(
        'RGB', ((xmax-xmin+1)*imgsize-1, (ymax-ymin+1)*imgsize-1))
    for xtile in range(xmin, xmax+1):
        for ytile in range(ymin,  ymax+1):
            threads.append(threading.Thread(target=get_img, args=(
                smurl, zoom, xtile, ytile, imgsize, imgsavepath)))
    for t in threads:
        t.setDaemon(True)
        t.start()
    for t in threads:
        t.join()
    threads.clear()
    return Cluster


def plot_map(plt, bounds, zoom='auto', style=0, printlog=False):
    '''
    Plot the basemap

    Parameters
    -------
    plt : matplotlib.pyplot
        Where to plot
    bounds : List
        The drawing boundary of the base map, [lon1,lat1,lon2,lat2]
        (WGS84 coordinate system), where lon1 and lat1 are the coordinates
        of the lower left corner and lon2 and lat2 are the coordinates of
        the upper right corner
    zoom : number
        The larger the magnification level of the base map, the longer the
        loading time. Generally, the range for a single city is between 12
        and 16
    printlog : bool
        Show log
    style : number
        The style of map basemap can be 1-10, as follows
    '''
    access_token = read_mapboxtoken()
    if access_token == '':
        style = 0
    imgsavepath = read_imgsavepath()
    if imgsavepath != '':
        try:
            import os
            os.listdir(imgsavepath)
        except Exception:
            warnings.warn('imgsavepath do not exist, your tile map will not save')
    else:
        warnings.warn('imgsavepath do not exist, your tile map will not save')
    lon1 = bounds[0]
    lat1 = bounds[1]
    lon2 = bounds[2]
    lat2 = bounds[3]
    if zoom == 'auto':
        zoom = 11-np.log(lon2-lon1)/np.log(2)
    zoom = min(18, int(zoom+0.5))
    a = getImageCluster(lon1, lat1, lon2-lon1,  lat2-lat1, zoom, style=style,
                        printlog=printlog, imgsavepath=imgsavepath,
                        access_token=access_token)
    x1, y1 = deg2num(lat1, lon1, zoom)
    x2, y2 = deg2num(lat2, lon2, zoom)
    x1, y1 = num2deg(x1, y1+1, zoom)
    x2, y2 = num2deg(x2+1, y2, zoom)
    plt.imshow(np.asarray(a), extent=(y1, y2, x1+0.00, x2+0.00))


def plotscale(ax, bounds, textcolor='k', textsize=8, compasssize=1,
              accuracy='auto', rect=[0.1, 0.1], unit="KM", style=1, **kwargs):
    '''
    Add compass and scale for a map

    Parameters
    -------
    bounds : List
        The drawing boundary of the base map, [lon1,lat1,lon2,lat2]
        (WGS84 coordinate system), where lon1 and lat1 are the coordinates
        of the lower left corner and lon2 and lat2 are the coordinates of
        the upper right corner
    textsize : number
        size of the text
    compasssize : number
        size of the compass
    accuracy : number
        Length of scale bar (m)
    unit : str
        ‘KM’,’km’,’M’,’m’, the scale units
    style : number
        1 or 2, the style of the scale
    rect : List
        The approximate position of the scale bar in the figure, such as
        [0.9,0.9], is in the upper right corner
    '''
    import math
    lon1 = bounds[0]
    lat1 = bounds[1]
    lon2 = bounds[2]
    lat2 = bounds[3]
    if accuracy == 'auto':
        accuracy = (int((lon2-lon1)/0.0003/1000+0.5)*1000)
    a, c = rect
    b = 1-a
    d = 1-c
    alon, alat = (b*lon1+a*lon2)/(a+b), (d*lat1+c*lat2)/(c+d)
    deltaLon = accuracy * 360 / \
        (2 * math.pi * 6371004 * math.cos((lat1 + lat2) * math.pi / 360))
    # add scale
    from shapely.geometry import Polygon
    import geopandas as gpd
    if style == 1:
        scale = gpd.GeoDataFrame({
            'color': [(0, 0, 0), (1, 1, 1), (0, 0, 0), (1, 1, 1)],
            'geometry':
            [Polygon([
                (alon, alat),
                (alon+deltaLon, alat),
                (alon+deltaLon, alat+deltaLon*0.4),
                (alon, alat+deltaLon*0.4)]),
             Polygon([
                 (alon+deltaLon, alat),
                 (alon+2*deltaLon, alat),
                 (alon+2 * deltaLon, alat+deltaLon*0.4),
                 (alon+deltaLon, alat+deltaLon*0.4)]),
             Polygon([
                 (alon+2*deltaLon, alat),
                 (alon+4*deltaLon, alat),
                 (alon+4 * deltaLon, alat+deltaLon*0.4),
                 (alon+2*deltaLon, alat+deltaLon*0.4)]),
             Polygon([
                 (alon+4*deltaLon, alat),
                 (alon+8*deltaLon, alat),
                 (alon+8 * deltaLon, alat+deltaLon*0.4),
                 (alon+4*deltaLon, alat+deltaLon*0.4)])
             ]})
        scale.plot(ax=ax, edgecolor=textcolor,
                   facecolor=scale['color'], lw=0.6, **kwargs)
        if (unit == 'KM') | (unit == 'km'):
            ax.text(alon+1*deltaLon, alat+deltaLon*0.5,
                    str(int(1*accuracy/1000)), color=textcolor,
                    fontsize=textsize, ha='center', va='bottom')
            ax.text(alon+2*deltaLon, alat+deltaLon*0.5,
                    str(int(2*accuracy/1000)), color=textcolor,
                    fontsize=textsize, ha='center', va='bottom')
            ax.text(alon+4*deltaLon, alat+deltaLon*0.5,
                    str(int(4*accuracy/1000)), color=textcolor,
                    fontsize=textsize, ha='center', va='bottom')
            ax.text(alon+8*deltaLon, alat+deltaLon*0.5,
                    str(int(8*accuracy/1000)), color=textcolor,
                    fontsize=textsize, ha='center', va='bottom')
            ax.text(alon+8.5*deltaLon, alat+deltaLon*0.5, unit,
                    color=textcolor, fontsize=textsize, ha='left',
                    va='top')
        if (unit == 'M') | (unit == 'm'):
            ax.text(alon+1*deltaLon, alat+deltaLon*0.5, str(int(1*accuracy)),
                    color=textcolor, fontsize=textsize,
                    ha='center', va='bottom')
            ax.text(alon+2*deltaLon, alat+deltaLon*0.5, str(int(2*accuracy)),
                    color=textcolor, fontsize=textsize,
                    ha='center', va='bottom')
            ax.text(alon+4*deltaLon, alat+deltaLon*0.5, str(int(4*accuracy)),
                    color=textcolor, fontsize=textsize,
                    ha='center', va='bottom')
            ax.text(alon+8*deltaLon, alat+deltaLon*0.5, str(int(8*accuracy)),
                    color=textcolor, fontsize=textsize,
                    ha='center', va='bottom')
            ax.text(alon+8.5*deltaLon, alat+deltaLon*0.5, unit,
                    color=textcolor, fontsize=textsize,
                    ha='left', va='top')
    if style == 2:
        scale = gpd.GeoDataFrame({
            'color': [(0, 0, 0), (1, 1, 1)],
            'geometry':
            [Polygon([(alon+deltaLon, alat),
                      (alon+4*deltaLon, alat),
                      (alon+4*deltaLon, alat+deltaLon*0.4),
                      (alon+deltaLon, alat+deltaLon*0.4)]),
                Polygon([(alon+4*deltaLon, alat),
                         (alon+8*deltaLon, alat),
                         (alon+8 * deltaLon, alat+deltaLon*0.4),
                         (alon+4*deltaLon, alat+deltaLon*0.4)])
             ]})
        scale.plot(ax=ax, edgecolor=textcolor,
                   facecolor=scale['color'], lw=0.6, **kwargs)
        if (unit == 'KM') | (unit == 'km'):
            ax.text(alon+4*deltaLon, alat+deltaLon*0.5,
                    str(int(4*accuracy/1000)),
                    color=textcolor, fontsize=textsize,
                    ha='center', va='bottom')
            ax.text(alon+8*deltaLon, alat+deltaLon*0.5,
                    str(int(8*accuracy/1000)),
                    color=textcolor, fontsize=textsize,
                    ha='center', va='bottom')
            ax.text(alon+8.5*deltaLon, alat+deltaLon*0.5, unit,
                    color=textcolor, fontsize=textsize,
                    ha='left', va='top')
        if (unit == 'M') | (unit == 'm'):
            ax.text(alon+4*deltaLon, alat+deltaLon*0.5,
                    str(int(4*accuracy)),
                    color=textcolor, fontsize=textsize,
                    ha='center', va='bottom')
            ax.text(alon+8*deltaLon, alat+deltaLon*0.5,
                    str(int(8*accuracy)),
                    color=textcolor, fontsize=textsize,
                    ha='center', va='bottom')
            ax.text(alon+8.5*deltaLon, alat+deltaLon*0.5, unit,
                    color=textcolor, fontsize=textsize,
                    ha='left', va='top')
    # add compass
    deltaLon = compasssize*deltaLon
    alon = alon-deltaLon
    compass = gpd.GeoDataFrame({
        'color': [(0, 0, 0), (1, 1, 1)],
        'geometry': [
            Polygon([[alon, alat],
                     [alon, alat+deltaLon],
                     [alon+1/2*deltaLon, alat-1/2*deltaLon]]),
            Polygon([[alon, alat],
                     [alon, alat+deltaLon],
                     [alon-1/2*deltaLon, alat-1/2*deltaLon]])]})
    compass.plot(ax=ax, edgecolor=textcolor,
                 facecolor=compass['color'], lw=0.6, **kwargs)
    ax.text(alon, alat+deltaLon, 'N', color=textcolor,
            fontsize=textsize, ha='center', va='bottom')



'''
BSD 3-Clause License

Copyright (c) 2021, Qing Yu
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its
   contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
'''

import numpy as np
x_pi = 3.14159265358979324 * 3000.0 / 180.0
pi = 3.1415926535897932384626
a = 6378245.0
ee = 0.00669342162296594323


def gcj02tobd09(lng, lat):
    """
    Convert coordinates from GCJ02 to BD09

    Parameters
    -------
    lng : Series or number
        Longitude
    lat : Series or number
        Latitude

    return
    -------
    lng : Series or number
        Longitude (Converted)
    lat : Series or number
        Latitude (Converted)
    """
    try:
        lng = lng.astype(float)
        lat = lat.astype(float)
    except Exception:
        lng = float(lng)
        lat = float(lat)
    z = np.sqrt(lng * lng + lat * lat) + 0.00002 * np.sin(lat * x_pi)
    theta = np.arctan2(lat, lng) + 0.000003 * np.cos(lng * x_pi)
    bd_lng = z * np.cos(theta) + 0.0065
    bd_lat = z * np.sin(theta) + 0.006
    return bd_lng, bd_lat


def bd09togcj02(bd_lon, bd_lat):
    """
    Convert coordinates from BD09 to GCJ02

    Parameters
    -------
    lng : Series or number
        Longitude
    lat : Series or number
        Latitude

    return
    -------
    lng : Series or number
        Longitude (Converted)
    lat : Series or number
        Latitude (Converted)
    """
    try:
        bd_lon = bd_lon.astype(float)
        bd_lat = bd_lat.astype(float)
    except Exception:
        bd_lon = float(bd_lon)
        bd_lat = float(bd_lat)
    x = bd_lon - 0.0065
    y = bd_lat - 0.006
    z = np.sqrt(x * x + y * y) - 0.00002 * np.sin(y * x_pi)
    theta = np.arctan2(y, x) - 0.000003 * np.cos(x * x_pi)
    gg_lng = z * np.cos(theta)
    gg_lat = z * np.sin(theta)
    return gg_lng, gg_lat


def wgs84togcj02(lng, lat):
    """
    Convert coordinates from WGS84 to GCJ02

    Parameters
    -------
    lng : Series or number
        Longitude
    lat : Series or number
        Latitude

    return
    -------
    lng : Series or number
        Longitude (Converted)
    lat : Series or number
        Latitude (Converted)
    """
    try:
        lng = lng.astype(float)
        lat = lat.astype(float)
    except Exception:
        lng = float(lng)
        lat = float(lat)
    dlat = transformlat(lng - 105.0, lat - 35.0)
    dlng = transformlng(lng - 105.0, lat - 35.0)
    radlat = lat / 180.0 * pi
    magic = np.sin(radlat)
    magic = 1 - ee * magic * magic
    sqrtmagic = np.sqrt(magic)
    dlat = (dlat * 180.0) / ((a * (1 - ee)) / (magic * sqrtmagic) * pi)
    dlng = (dlng * 180.0) / (a / sqrtmagic * np.cos(radlat) * pi)
    mglat = lat + dlat
    mglng = lng + dlng
    return mglng, mglat


def gcj02towgs84(lng, lat):
    """
    Convert coordinates from GCJ02 to WGS84

    Parameters
    -------
    lng : Series or number
        Longitude
    lat : Series or number
        Latitude

    return
    -------
    lng : Series or number
        Longitude (Converted)
    lat : Series or number
        Latitude (Converted)
    """
    try:
        lng = lng.astype(float)
        lat = lat.astype(float)
    except Exception:
        lng = float(lng)
        lat = float(lat)
    dlat = transformlat(lng - 105.0, lat - 35.0)
    dlng = transformlng(lng - 105.0, lat - 35.0)
    radlat = lat / 180.0 * pi
    magic = np.sin(radlat)
    magic = 1 - ee * magic * magic
    sqrtmagic = np.sqrt(magic)
    dlat = (dlat * 180.0) / ((a * (1 - ee)) / (magic * sqrtmagic) * pi)
    dlng = (dlng * 180.0) / (a / sqrtmagic * np.cos(radlat) * pi)
    mglat = lat + dlat
    mglng = lng + dlng
    return lng * 2 - mglng, lat * 2 - mglat


def wgs84tobd09(lon, lat):
    """
    Convert coordinates from WGS84 to BD09

    Parameters
    -------
    lng : Series or number
        Longitude
    lat : Series or number
        Latitude

    return
    -------
    lng : Series or number
        Longitude (Converted)
    lat : Series or number
        Latitude (Converted)
    """
    try:
        lon = lon.astype(float)
        lat = lat.astype(float)
    except Exception:
        lon = float(lon)
        lat = float(lat)
    lon, lat = wgs84togcj02(lon, lat)
    lon, lat = gcj02tobd09(lon, lat)
    return lon, lat


def bd09towgs84(lon, lat):
    """
    Convert coordinates from BD09 to WGS84

    Parameters
    -------
    lng : Series or number
        Longitude
    lat : Series or number
        Latitude

    return
    -------
    lng : Series or number
        Longitude (Converted)
    lat : Series or number
        Latitude (Converted)
    """
    try:
        lon = lon.astype(float)
        lat = lat.astype(float)
    except Exception:
        lon = float(lon)
        lat = float(lat)
    lon, lat = bd09togcj02(lon, lat)
    lon, lat = gcj02towgs84(lon, lat)
    return lon, lat


def bd09mctobd09(x, y):
    """
    Convert coordinates from BD09MC to BD09

    Parameters
    -------
    x : Series or number
        x coordinates
    y : Series or number
        y coordinates

    return
    -------
    lng : Series or number
        Longitude (Converted)
    lat : Series or number
        Latitude (Converted)
    """
    MCBAND = [12890594.86, 8362377.87, 5591021, 3481989.83, 1678043.12, 0]
    MC2LL = [
        [1.410526172116255e-8, 0.00000898305509648872, -1.9939833816331,
         200.9824383106796, -187.2403703815547, 91.6087516669843,
         -23.38765649603339, 2.57121317296198, -0.03801003308653, 17337981.2],
        [-7.435856389565537e-9, 0.000008983055097726239, -0.78625201886289,
         96.32687599759846, -1.85204757529826, -59.36935905485877,
         47.40033549296737, -16.50741931063887, 2.28786674699375, 10260144.86],
        [-3.030883460898826e-8, 0.00000898305509983578, 0.30071316287616,
         59.74293618442277, 7.357984074871, -25.38371002664745,
         13.45380521110908, -3.29883767235584, 0.32710905363475, 6856817.37],
        [-1.981981304930552e-8, 0.000008983055099779535, 0.03278182852591,
         40.31678527705744, 0.65659298677277, -4.44255534477492,
         0.85341911805263, 0.12923347998204, -0.04625736007561, 4482777.06],
        [3.09191371068437e-9, 0.000008983055096812155, 0.00006995724062,
         23.10934304144901, -0.00023663490511, -0.6321817810242,
         -0.00663494467273, 0.03430082397953, -0.00466043876332, 2555164.4],
        [2.890871144776878e-9, 0.000008983055095805407, -3.068298e-8,
         7.47137025468032, -0.00000353937994, -0.02145144861037,
         -0.00001234426596, 0.00010322952773, -0.00000323890364, 826088.5]
    ]
    y1 = y.iloc[0]
    for cD in range(len(MCBAND)):
        if y1 >= MCBAND[cD]:
            cE = MC2LL[cD]
            break
    cD = cE
    T = cD[0] + cD[1] * np.abs(x)
    cB = np.abs(y) / cD[9]
    cE = cD[2] + cD[3] * cB + cD[4] * cB * cB +\
        cD[5] * cB * cB * cB + cD[6] * cB * cB * cB * cB +\
        cD[7] * cB * cB * cB * cB * cB +\
        cD[8] * cB * cB * cB * cB * cB * cB
    return T, cE


def transformlat(lng, lat):
    ret = -100.0 + 2.0 * lng + 3.0 * lat + 0.2 * lat * lat + \
        0.1 * lng * lat + 0.2 * np.sqrt(np.fabs(lng))
    ret += (20.0 * np.sin(6.0 * lng * pi) + 20.0 *
            np.sin(2.0 * lng * pi)) * 2.0 / 3.0
    ret += (20.0 * np.sin(lat * pi) + 40.0 *
            np.sin(lat / 3.0 * pi)) * 2.0 / 3.0
    ret += (160.0 * np.sin(lat / 12.0 * pi) + 320 *
            np.sin(lat * pi / 30.0)) * 2.0 / 3.0
    return ret


def transformlng(lng, lat):
    import numpy as np
    ret = 300.0 + lng + 2.0 * lat + 0.1 * lng * lng + \
        0.1 * lng * lat + 0.1 * np.sqrt(np.abs(lng))
    ret += (20.0 * np.sin(6.0 * lng * pi) + 20.0 *
            np.sin(2.0 * lng * pi)) * 2.0 / 3.0
    ret += (20.0 * np.sin(lng * pi) + 40.0 *
            np.sin(lng / 3.0 * pi)) * 2.0 / 3.0
    ret += (150.0 * np.sin(lng / 12.0 * pi) + 300.0 *
            np.sin(lng / 30.0 * pi)) * 2.0 / 3.0
    return ret


def getdistance(lon1, lat1, lon2, lat2):
    '''
    Input the origin/destination location in the sequence of [lon1,
    lat1, lon2, lat2] (in decimal) from DataFrame. The output is the
    distance (m).

    Parameters
    -------
    lon1 : Series or number
        Start longitude
    lat1 : Series or number
        Start latitude
    lon2 : Series or number
        End longitude
    lat2 : Series or number
        End latitude

    return
    -------
    distance : Series or number
        The distance
    '''
    try:
        lon1 = lon1.astype(float)
        lat1 = lat1.astype(float)
        lon2 = lon2.astype(float)
        lat2 = lat2.astype(float)
    except Exception:
        lon1 = float(lon1)
        lat1 = float(lat1)
        lon2 = float(lon2)
        lat2 = float(lat2)
    lon1, lat1, lon2, lat2 = map(lambda r: r*pi/180, [lon1, lat1, lon2, lat2])
    dlon = lon2 - lon1
    dlat = lat2 - lat1
    a = np.sin(dlat/2)**2 + np.cos(lat1) * np.cos(lat2) * np.sin(dlon/2)**2
    c = 2 * np.arcsin(a**0.5)
    r = 6371  # 地球平均半径，单位为公里
    return c * r * 1000


def transform_shape(gdf, method):
    '''
    Convert coordinates of all data.
    The input is the geographic elements’ DataFrame.

    Parameters
    -------
    gdf : GeoDataFrame
        Geographic elements
    method : function
        The coordinate converting function

    return
    -------
    gdf : GeoDataFrame
        The result of converting
    '''
    from shapely.ops import transform
    gdf1 = gdf.copy()
    gdf1['geometry'] = gdf1['geometry'].apply(lambda r: transform(method, r))
    return gdf1



'''
BSD 3-Clause License

Copyright (c) 2021, Qing Yu
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its
   contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
'''

import pandas as pd
import numpy as np
from shapely.geometry import LineString


def split_subwayline(line, stop):
    '''
    To slice the metro line with metro stations to obtain metro section
    information (This step is useful in subway passenger flow visualization)

    Parameters
    -------
    line : GeoDataFrame
        Bus/metro lines
    stop : GeoDataFrame
        Bus/metro stations

    Returns
    -------
    metro_line_splited : GeoDataFrame
        Generated section line shape
    '''
    def getline(r2, line_geometry):
        ls = []
        if r2['o_project'] <= r2['d_project']:
            tmp1 = np.linspace(r2['o_project'], r2['d_project'], 10)
        if r2['o_project'] > r2['d_project']:
            tmp1 = np.linspace(
                r2['o_project']-line_geometry.length, r2['d_project'], 10)
            tmp1[tmp1 < 0] = tmp1[tmp1 < 0]+line_geometry.length
        for j in tmp1:
            ls.append(line_geometry.interpolate(j))
        return LineString(ls)
    lss = []
    for k in range(len(line)):
        r = line.iloc[k]
        line_geometry = r['geometry']
        tmp = stop[stop['linename'] == r['linename']].copy()
        for i in tmp.columns:
            tmp[i+'1'] = tmp[i].shift(-1)
        tmp = tmp.iloc[:-1]
        tmp = tmp[['stationnames', 'stationnames1',
                   'geometry', 'geometry1', 'linename', 'line']]
        tmp['o_project'] = tmp['geometry'].apply(
            r['geometry'].project)
        tmp['d_project'] = tmp['geometry1'].apply(
            r['geometry'].project)
        tmp['geometry'] = tmp.apply(
            lambda r2: getline(r2, line_geometry), axis=1)
        lss.append(tmp)
    metro_line_splited = pd.concat(lss).drop('geometry1', axis=1)
    metro_line_splited.crs = 'epsg:4326'
    metro_line_splited['length'] = metro_line_splited.to_crs(epsg=3857).length
    metro_line_splited = metro_line_splited.drop(
        ['o_project', 'd_project'], axis=1)
    return metro_line_splited


def metro_network(line, stop, transfertime=5, nxgraph=True):
    '''
    Inputting the metro station data and outputting the network topology
    model. The graph generated relies on NetworkX. The travel time is consist of:
    operation time between stations + stop time at each station + transfer time

    Parameters
    -------

    line : GeoDataFrame
        Lines. Should have `line` column to store line name `speed` column to
        store metro speed and `stoptime` column to store stop time at each
        station.
    stop : GeoDataFrame
        Bus/metro stations
    transfertime : number
        Travel time per transfer
    nxgraph : bool
        Default True, if True then output the network G constructed by
        NetworkX, if False then output the edges1(line section),
        edge2(station transfer), and the node of the network

    Returns
    -------

    When the nxgraph parameter is True
    ================================================================
    G : networkx.classes.graph.Graph
        Network G built by networkx.

    When the nxgraph parameter is False (This is for detail design of the network)
    ================================================================
    edge1 : DataFrame
        Network edge for line section.
    edge2 : DataFrame
        Network edge for transfering.
    node : List
        Network nodes.
    '''
    # Obtain edge1: Network edge for line section.
    linestop = stop.copy()
    if ('speed' not in line.columns) | ('stoptime' not in line.columns):
        raise ValueError(
            'Lines should have `line` column to store line name,'
            '`speed` column to store metro speed and'
            '`stoptime` column to store stop time at each station'
        )
    for i in linestop.columns:
        linestop[i+'1'] = linestop[i].shift(-1)
    linestop = linestop[linestop['linename'] == linestop['linename1']].copy()
    linestop = linestop.rename(
        columns={'stationnames': 'ostop', 'stationnames1': 'dstop'})
    linestop['ostation'] = linestop['line']+linestop['ostop']
    linestop['dstation'] = linestop['line']+linestop['dstop']
    edge1 = linestop[['ostation', 'dstation']].copy()

    # calculate travel time for edge1
    # calculate distance
    metrolinesplit = split_subwayline(line, stop)
    metrolinesplit['ostation'] = metrolinesplit['line'] + \
        metrolinesplit['stationnames']
    metrolinesplit['dstation'] = metrolinesplit['line'] + \
        metrolinesplit['stationnames1']
    metrolinesplit = metrolinesplit[['ostation', 'dstation', 'line', 'length']]
    edge1 = pd.merge(edge1, metrolinesplit, how='left')
    edge1 = pd.merge(edge1, line[['line', 'speed', 'stoptime']])

    # calculate duration
    edge1['duration'] = 60*(edge1['length']/1000) / \
        edge1['speed']+edge1['stoptime']
    edge1 = edge1[['ostation', 'dstation', 'duration']].drop_duplicates(
        subset=['ostation', 'dstation'])

    # Obtain edge2: Network edge for transfering.
    linestop = stop.copy()
    linestop['station'] = linestop['line'] + linestop['stationnames']
    tmp = linestop.groupby(['stationnames'])[
        'linename'].count().rename('count').reset_index()
    tmp = pd.merge(linestop, tmp[tmp['count'] > 2]
                   ['stationnames'], on='stationnames')
    tmp = tmp[['stationnames', 'line', 'station']].drop_duplicates()
    tmp = pd.merge(tmp, tmp, on='stationnames')

    edge2 = tmp[tmp['line_x'] != tmp['line_y']][['station_x', 'station_y']]
    # All transfer time are set as the same, export `edge2` for further degign
    edge2['duration'] = transfertime
    edge2.columns = edge1.columns
    edge = edge1.append(edge2)
    node = list(edge['ostation'].drop_duplicates())
    if nxgraph:
        import networkx as nx
        G = nx.Graph()
        G.add_nodes_from(node)
        G.add_weighted_edges_from(edge.values)
        return G
    else:
        return edge1, edge2, node


def get_shortest_path(G, stop, ostation, dstation):
    '''
    Obtain the travel path of shortest path from the metro nextwork

    Parameters
    -------

    G : networkx.classes.graph.Graph
        metro network
    stop : DataFrame
        metro stop dataframe
    ostation : str
        O station name
    dstation : str
        D station name


    Returns
    -------
    path : list
        travel path: list of station names
    '''
    import networkx as nx
    o = stop[stop['stationnames'] == ostation]['line'].iloc[0]+ostation
    d = stop[stop['stationnames'] == dstation]['line'].iloc[0]+dstation
    return nx.shortest_path(G, source=o, target=d, weight='weight')


def get_k_shortest_paths(G, stop, ostation, dstation, k):
    '''
    Obtain the k th shortest paths from the metro nextwork

    Parameters
    -------

    G : networkx.classes.graph.Graph
        metro network
    stop : DataFrame
        metro stop dataframe
    ostation : str
        O station name
    dstation : str
        D station name
    k : int
        the k th shortest paths


    Returns
    -------
    paths : list
        travel path: list of travel paths
    '''
    from itertools import islice
    import networkx as nx
    o = stop[stop['stationnames'] == ostation]['line'].iloc[0]+ostation
    d = stop[stop['stationnames'] == dstation]['line'].iloc[0]+dstation
    return list(
        islice(nx.shortest_simple_paths(G, o, d, weight='weight'), k)
    )


def get_path_traveltime(G, path):
    '''
    Obtain the travel time of the path

    Parameters
    -------

    G : networkx.classes.graph.Graph
        metro network
    path : list
        list of stationnames

    Returns
    -------
    traveltime : float
        travel time of the path
    '''
    traveltime = 0
    for i in range(len(path)-1):
        traveltime += G.get_edge_data(path[i], path[i+1])['weight']
    return traveltime
